#define __OXL_C

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>

#include "oxl.h"

/**************************************/

void oxl_close();

void oxl_fatal(int i)
{
	fprintf(stderr,"guru meditation %.8x\n",i);
	oxl_close();
	exit(i);
}
void oxl_quit()
{
	oxl.quitsigns++;

	signal(SIGINT,&oxl_fatal);
	signal(SIGQUIT,&oxl_fatal);
}
int oxl_quitcount()
{
	return oxl.quitsigns;
}
void oxl_regdestr(void(*func)())
{
	static int curr=0;
	oxl.destr[curr++]=func;
	oxl.destr[curr]=NULL;
}
void oxl_close()
{
	int i=0;
	OXLMSG("closing oxl\n");
	for(;;){
		if(oxl.destr[i]==NULL)break;
		else oxl.destr[i++]();
	}
}
int oxl_testquit()
{
	static int quitreqs=0;

	quitreqs+=oxl.drv.testquit()+oxl.quitsigns;
	if(quitreqs){oxl_close();exit(1);}
}

/***/ /* getopts stuff */

void rmarg(int*argc,char**argv,int n)
{
	while(n<*argc){argv[n]=argv[n+1];n++;}
	(*argc)--;
}
int findarg(int*argc,char**argv,char*key)
{
	int i=1;
	for(;i<*argc;i++)if(!strcmp(key,argv[i])){
		rmarg(argc,argv,i);return i;}
	return 0;
}

/**************************************/

int oxl_evalmode(w,h,d,vw,vh,vd,max)
int w,h,d,vw,vh,vd,*max;
{
	int pts = (abs(w-vw)<<4)/w + (abs(h-vh)<<4)/h;
	if((w<vw)||(h<vh))pts+=4;

	if(vw==2*w){
		if(vh==2*h)pts=6;else pts=10;
	}
	if(!vd)return 0; /* ? */

	if(d!=vd){
		pts+=6;
		if(vd<d){
			pts+=4;
			if(vd!=8)pts+=8;
		}
	}
#ifdef DEBUG	
	fprintf(stderr,"%ix%ix%i ==> %ix%ix%i gets %i\n",
		w,h,d,vw,vh,vd,pts);
#endif
	if(pts<*max){
		*max=pts;
		oxl.buf.vw=vw;
		oxl.buf.vh=vh;
		oxl.buf.vd=vd;
		return 1;
	}
	return 0;
}

/*** timer api ************************/

int oxlunix_timer_tod()
{
	struct timeval tod;
	gettimeofday(&tod,NULL);
	return (tod.tv_sec*oxl.timerhz)+(tod.tv_usec*oxl.timerhz/1000000);
}
int oxlunix_settimer()
{
	oxl.drv.timer=oxlunix_timer_tod;
	/* we should include support for ITIMERs */
}
int oxl_timer(int hz)
{
	static int base;int v;
	if(hz){base=-1;oxl.timerhz=hz;}
	if(oxl.drv.timer==NULL)oxlunix_settimer();
	v=oxl.drv.timer();
	#ifdef DEBUG
	fprintf(stderr,"timer:%i (base %i)\n",v,base);
	#endif
	if(base==-1)base=v;
	return v-base;
}

/*** initializer **********************/

#define FULLINIT "xvgao"
void oxl_initdrvs(int*argc,char**argv,char*s)
{
	while(*s){
		switch((int)*s|32){
#ifdef SUPPORT_X11
			case('x'):oxlX11_init(argc,argv);break;
#endif
#ifdef SUPPORT_VGA
			case('v'):oxlvga_init(argc,argv);break;
#endif
#ifdef SUPPORT_GGI
			case('g'):oxlggi_init(argc,argv);break;
#endif
#ifdef SUPPORT_AA
			case('a'):oxlaa_init(argc,argv);break;
#endif
#ifdef SUPPORT_OSS
			case('o'):
				if(!oxl.nosnd)
				oxloss_init(argc,argv);
				break;
#endif
		}
		s++;}
}

void oxl_init(int*argc,char**argv)
{
	int i;
	oxl.help=findarg(argc,argv,"-h")|findarg(argc,argv,"--help");
	if(oxl.help)
		OXLMSG("-d <drivers " FULLINIT ">\n-z <width> <height> (zoom to)\n"
		       "-c <width> <height> (calc to)\n"
		       "-f <audiofreq>\n-nosnd\n");
	else{
		OXLMSG("ocsalib: -h for help\n");

		/* pre-drvinit settings */
		oxl.quitsigns=0;
		oxl.drv.testquit=oxl_quitcount;
		oxl.destr[0]=NULL;
		oxl.drv.audio_send=NULL;

		/* checkopts */		
		oxl.nosnd=findarg(argc,argv,"-nosnd");
		for(;;){
		i=findarg(argc,argv,"-d");
		if(i)oxl_initdrvs(argc,argv,argv[i]);else break;
		}
		/* resolutions */
		if((i=findarg(argc,argv,"-z"))){
			oxl.opt.zoom=OXL_ZOOM_USERZ;
			oxl.buf.vw=atoi(argv[i]);
			oxl.buf.vh=atoi(argv[i+1]);
		}else oxl.opt.zoom=0;
		if((i=findarg(argc,argv,"-c"))){
			oxl.opt.zoom|=OXL_ZOOM_USERC;
			oxl.buf.uw=atoi(argv[i]);
			oxl.buf.uh=atoi(argv[i+1]);
		}
		oxl.au.hz=(i=findarg(argc,argv,"-f"))
			?atoi(argv[i]):OXLAU_HZDEFAULT;
		oxl.au.bits=OXLAU_BITSDEFAULT;
		oxl.au.chs=OXLAU_CHSDEFAULT;
	}

	oxl_initdrvs(argc,argv,FULLINIT);
	if(oxl.help)exit(0);

	/* post-drvinit settings */

		oxl.drv.timer=NULL; /* no olkoot tuossa vhn aikaa */
		oxl.drv.audio_check=NULL;

	oxl.inited=1;
 	signal(SIGTERM,&oxl_fatal);
	signal(SIGINT,&oxl_quit);
	signal(SIGQUIT,&oxl_quit);
	signal(SIGFPE,&oxl_fatal);
	signal(SIGSEGV,&oxl_fatal);
}

/*** audio api ************************/

void oxl_sendaudio(void*buffer,int length)
{
	static struct{
		void*data; int lgt;
	}tmpbuf={NULL,0};

	if(oxl.drv.audio_send!=NULL){

		if(length>tmpbuf.lgt){
		if(tmpbuf.lgt)free(tmpbuf.data);
		tmpbuf.data=malloc((tmpbuf.lgt=length)*AUBYPS);}

		/* oxl.au.outmode should select scalers&clippers */
		oxl_mixer_clip32(buffer,length);
		if(oxl.au.bits==8)
		oxl_mixer_scale8u(tmpbuf.data,buffer,length);else
		oxl_mixer_scale16u(tmpbuf.data,buffer,length);

		oxl.drv.audio_send(tmpbuf.data,length*AUBYPS);
	}
} /* OXL_RAWDATA ? */

void oxl_play(void(*player)())
{
	if(oxl.drv.audio_send!=NULL){

	player();

	if(oxl.drv.audio_play!=NULL)
	oxl.drv.audio_play(player);else
		if(fork())for(;;)oxl.drv.audio_play(player);
	}
}

/*** video api ************************/

extern void oxl_scaler_setpalette(int*pal,int colorcount);

void oxl_setpalette(int*pal,int colorcount)
{
   int i,defpal=0;

	#ifdef SUPPORT_NULLPALETTE
   if(pal==NULL){
      pal=malloc(colorcount*3*sizeof(int));
      for(i=0;i<colorcount*3;i++){
      	int tmp=(i*256)/(3*colorcount);if(tmp>255)tmp=255;
      	pal[i]=tmp;}
      defpal=1;
   }
	#endif

	oxl.drv.video_setpalette(pal,colorcount);

	#ifdef SUPPORT_NULLPALETTE
   if(defpal)free(pal);
   #endif
}

void* oxl_video(int w,int h,int d)
{
	static int readied=0;

	oxl.buf.w=w;oxl.buf.h=h;oxl.buf.d=d;

	if(readied){
		if(oxl.use_scaler)free(oxl.buf.data);
	}else{
		oxl.buf.vd=d;
		if(!(oxl.opt.zoom&OXL_ZOOM_USERZ)){
			oxl.buf.vw=w;
			oxl.buf.vh=h;}
		if(!oxl.inited)oxl_init(0,NULL);
	if(oxl.drv.video_ready==NULL)
	{
		OXLMSG("unable to video_init\n");
		exit(1);
	}
	oxl.drv.video_ready();
	}

	if( (oxl.buf.w==oxl.buf.vw)
	  &&(oxl.buf.h==oxl.buf.vh)
	  &&(oxl.buf.d==oxl.buf.vd)){
		oxl.use_scaler=0;
		oxl.buf.data=oxl.buf.vdata;
	}else
	{
		oxl.use_scaler=1;
		oxl_scaler_ready();
		
		#ifdef SUPPORT_8TOHIGHER
		if(oxl.buf.vd!=8)oxl.drv.video_setpalette=oxl_scaler_setpalette;
		#endif

	#ifdef DEBUG
		fprintf(stderr,"alloc data buf: %i %i\n",oxl.buf.w,oxl.buf.h);
	#endif
		oxl.buf.data=malloc(BUFFSIZE(oxl.buf.w,oxl.buf.h,oxl.buf.d));
	};

	#ifdef DEBUG
	fprintf(stderr,"%i/%i %i/%i %i/%i\n",
		oxl.buf.w,oxl.buf.vw,oxl.buf.h,oxl.buf.vh,oxl.buf.d,oxl.buf.vd);
	#endif

	if(!readied){
		if(!(oxl.opt.zoom&OXL_ZOOM_USERC)){
			oxl.buf.uw=oxl.buf.vw;
			oxl.buf.uh=oxl.buf.vh;
	}	}

	readied++;
	return oxl.buf.data;
}

void oxl_blit(void*buf)
{
	if(oxl.use_scaler)oxl_scaler_do();

	if(!oxl.directvideo)oxl.drv.video_blit();
		/* else oxl.drv.video_pageflip(); */

	if(oxl.drv.audio_check!=NULL)oxl.drv.audio_check();

	oxl_testquit();
}
